@@ -10,6 +10,7 @@ import threading
 
 import gobject
 import gtk
+import pango
 from gettext import gettext as _
 from gettext import ngettext
 
@@ -18,9 +19,9 @@ import game_selector
 import gsudoku
 import printing
 import saver
-import sudoku_generator_gui
 import sudoku_maker
 import timer
+import tracker_info
 from defaults import (APPNAME, APPNAME_SHORT, AUTHORS, COPYRIGHT, DESCRIPTION, DOMAIN, 
         IMAGE_DIR, LICENSE, MIN_NEW_PUZZLES, UI_DIR, VERSION, WEBSITE, WEBSITE_LABEL)
 from gtk_goodies import gconf_wrapper, Undo, dialog_extras
@@ -47,18 +48,18 @@ def inactivate_new_game_etc (fun):
     def inactivate_new_game_etc_ (ui, *args, **kwargs):
         paths = [
             '/MenuBar/Game/New',
+            '/MenuBar/Game/Reset',
+            '/MenuBar/Game/PuzzleInfo',
             '/MenuBar/Game/Print',
             # undo/redo is handled elsewhere as it can't simply be turned on/off.
-            '/MenuBar/Edit/Clear',
-            '/MenuBar/Edit/ClearNotes',
-            '/MenuBar/View/ToggleToolbar',
+            '/MenuBar/Settings/ToggleToolbar',
+            '/MenuBar/Settings/ToggleHighlight',
+            '/MenuBar/Settings/AlwaysShowPossible',
+            '/MenuBar/Settings/ShowImpossibleImplications',
             '/MenuBar/Tools/ShowPossible',
-            '/MenuBar/Tools/AutofillCurrentSquare',
-            '/MenuBar/Tools/Autofill',
-            '/MenuBar/Tools/AlwaysShowPossible',
-            '/MenuBar/Tools/ShowImpossibleImplications',
+            '/MenuBar/Tools/ClearTopNotes',
+            '/MenuBar/Tools/ClearBottomNotes',
             '/MenuBar/Tools/Tracker',
-            '/MenuBar/Game/PuzzleInfo',
             ]
         for p in paths:
             action = ui.uimanager.get_action(p)
@@ -85,6 +86,10 @@ class UI (gconf_wrapper.GConfWrapper):
     <menubar name="MenuBar">
       <menu name="Game" action="Game">
         <menuitem action="New"/>
+        <menuitem action="Reset"/>
+        <separator/>
+        <menuitem action="Undo"/>
+        <menuitem action="Redo"/>
         <separator/>
         <menuitem action="PuzzleInfo"/>
         <separator/>
@@ -93,32 +98,22 @@ class UI (gconf_wrapper.GConfWrapper):
         <separator/>
         <menuitem action="Close"/>
       </menu>
-      <menu action="Edit">
-        <menuitem action="Undo"/>
-        <menuitem action="Redo"/>
-        <separator/>
-        <menuitem action="Clear"/>
-        <menuitem action="ClearNotes"/>
-      </menu>
-      <menu action="View">
+      <menu action="Settings">
         <menuitem action="FullScreen"/>
-        <separator/>
         <menuitem action="ToggleToolbar"/>
+        <separator/>
         <menuitem action="ToggleHighlight"/>
+        <menuitem action="AlwaysShowPossible"/>
+        <menuitem action="ShowImpossibleImplications"/>
       </menu>
       <menu action="Tools">
         <menuitem action="ShowPossible"/>
-        <menuitem action="AutofillCurrentSquare"/>
-        <menuitem action="Autofill"/>
         <separator/>
-        <menuitem action="AlwaysShowPossible"/>
-        <menuitem action="ShowImpossibleImplications"/>
-        <separator/>
-        <menuitem action="Generator"/>
-        <menuitem action="BackgroundGenerator"/>
+        <menuitem action="ClearTopNotes"/>
+        <menuitem action="ClearBottomNotes"/>
         <separator/>
         <menuitem action="Tracker"/>
-        </menu>
+      </menu>
       <menu action="Help">
         <menuitem action="ShowHelp"/>
         <menuitem action="About"/>
@@ -126,15 +121,12 @@ class UI (gconf_wrapper.GConfWrapper):
     </menubar>
     <toolbar name="Toolbar">
       <toolitem action="New"/>
-      <toolitem action="Print"/>
       <separator/>
       <toolitem action="Undo"/>
       <toolitem action="Redo"/>
       <separator/>
       <toolitem action="ShowPossible"/>
-      <toolitem action="AutofillCurrentSquare"/>
       <separator/>
-      <toolitem action="ToggleHighlight"/>
       <toolitem action="Tracker"/>
     </toolbar>
     </ui>'''
@@ -164,6 +156,7 @@ class UI (gconf_wrapper.GConfWrapper):
                                             )
         self.setup_gui()
         self.timer = timer.ActiveTimer(self.w)
+        self.gsd.set_timer(self.timer)
         self.won = False
         # add the accelerator group to our toplevel window
         self.worker_connections = []
@@ -181,25 +174,21 @@ class UI (gconf_wrapper.GConfWrapper):
                 self.quit = True
             else:
                 self.quit = False
-                # Generate puzzles in background...
-                if self.gconf['generate_puzzles_in_background']:
-                    gobject.timeout_add_seconds(1, lambda *args: self.start_worker_thread() and True)
 
+        # Generate puzzles in background...
+        gobject.timeout_add_seconds(1, lambda *args: self.start_worker_thread() and True)
 
     @inactivate_new_game_etc
     def select_game (self):
         self.tb.hide()
-        self.update_statusbar()
         choice = game_selector.NewOrSavedGameSelector().run_swallowed_dialog(self.swallower)
         if not choice:
             return True
         self.timer.start_timing()
         if choice[0] == game_selector.NewOrSavedGameSelector.NEW_GAME:
             self.gsd.change_grid(choice[1], 9)
-            self.update_statusbar()
         if choice[0] == game_selector.NewOrSavedGameSelector.SAVED_GAME:
             saver.open_game(self, choice[1])
-            self.update_statusbar()
         if self.gconf['show_toolbar']:
             self.tb.show()
         if self.gconf['always_show_hints']:
@@ -216,6 +205,7 @@ class UI (gconf_wrapper.GConfWrapper):
         self.initialize_prefs()
         self.setup_main_window()
         self.gsd = gsudoku.SudokuGameDisplay()
+        self.gsd.set_parent_for(self.w)
         self.gsd.connect('puzzle-finished', self.you_win_callback)
         self.setup_color()
         self.setup_actions()
@@ -239,95 +229,80 @@ class UI (gconf_wrapper.GConfWrapper):
         self.main_actions = gtk.ActionGroup('MainActions')
         self.main_actions.add_actions([
             ('Game', None, _('_Game')),
-            ('New', gtk.STOCK_NEW, None,
-             '<Control>n', _('New game'), self.new_cb),
-            ('Print', gtk.STOCK_PRINT, None,
-             None, _('Print current game'), self.print_game),
-            ('PrintMany', gtk.STOCK_PRINT, _('Print _Multiple Sudokus'),
-             None, _('Print more than one sudoku at a time.'), self.print_multiple_games),
-            ('Close', gtk.STOCK_CLOSE, None, '<Control>w',
-             _('Close Sudoku'), self.quit_cb),
+            ('New', gtk.STOCK_NEW, None, '<Control>n', _('New game'), self.new_cb),
+            ('Reset', gtk.STOCK_CLEAR, _('_Reset'), '<Control>b',
+             _("Reset current grid(do-over)"), self.game_reset_cb),
+            ('Undo', gtk.STOCK_UNDO, _('_Undo'), '<Control>z',
+             _('Undo last action'), self.stop_dancer),
+            ('Redo', gtk.STOCK_REDO, _('_Redo'), '<Shift><Control>z',
+             _('Redo last action')),
+            ('PuzzleInfo', gtk.STOCK_ABOUT, _('Puzzle _Statistics...'), None,
+             _('Show statistics about current puzzle'), self.show_info_cb),
+            ('Print', gtk.STOCK_PRINT, _('_Print...'), '<Control>p', _('Print current game'), self.print_game),
+            ('PrintMany', gtk.STOCK_PRINT, _('Print _Multiple Sudokus...'), None,
+             _('Print more than one sudoku at a time.'), self.print_multiple_games),
+            ('Close', gtk.STOCK_CLOSE, None, '<Control>w', _('Close Sudoku'), self.quit_cb),
+            ('Settings', None, _('_Settings')),
+            ('FullScreen', gtk.STOCK_FULLSCREEN, None, 'F11', None, self.full_screen_cb),
             ('Tools', None, _('_Tools')),
-            ('View', None, _('_View')),
-            ('ShowPossible', gtk.STOCK_DIALOG_INFO, _('_Hint'),
-             '<Control>h',
-             _('Show which numbers could go in the current square.'),
-             self.show_hint_cb),
-            ('AutofillCurrentSquare', gtk.STOCK_APPLY, _('_Fill'), '<Control>f',
-             _('Automatically fill in the current square if possible.'),
-             self.auto_fill_current_square_cb),
-            ('Autofill', gtk.STOCK_REFRESH, _('Fill _all squares'), '<Control>a',
-             _('Automatically fill in all squares for which there is only one valid value.'),
-             self.auto_fill_cb),
-            ('FullScreen', gtk.STOCK_FULLSCREEN, None,
-             'F11', None, self.full_screen_cb),
-            ('Generator', None, _('_Generate new puzzles'), None, _('Generate new puzzles.'),
-              self.generate_puzzle_gui, ),
-            ('PuzzleInfo', gtk.STOCK_ABOUT, _('Puzzle _Statistics'),
-             None, _('Show statistics about current puzzle'),
-             self.show_info_cb),
-            ('Help', None, _('_Help'),
-             None, None, None),
-            ('About', gtk.STOCK_ABOUT, None,
-             None, None, self.show_about),
-            ('ShowHelp', gtk.STOCK_HELP, _('_Contents'),
-             'F1', None, self.show_help),
+            ('ShowPossible', gtk.STOCK_DIALOG_INFO, _('_Hint'), '<Control>h',
+             _('Show which numbers could go in the current square.'), self.show_hint_cb),
+            ('ClearTopNotes', None, _('Clear _Top Notes'), '<Control>j',
+             _("Clear all of the top notes"), self.clear_top_notes_cb),
+            ('ClearBottomNotes', None, _('Clear _Bottom Notes'), '<Control>k',
+             _("Clear all of the bottom notes"), self.clear_bottom_notes_cb),
+            ('Help', None, _('_Help'), None, None, None),
+            ('ShowHelp', gtk.STOCK_HELP, _('_Contents'), 'F1', None, self.show_help),
+            ('About', gtk.STOCK_ABOUT, None, None, None, self.show_about),
             ])
         self.main_actions.add_toggle_actions([
             ('AlwaysShowPossible',
              None,
-             _('_Always show hint'),
+             _('Show _Possible Numbers'),
              None,
              _('Always show possible numbers in a square'),
              self.auto_hint_cb),
             ('ShowImpossibleImplications',
              None,
-             _('Warn about _unfillable squares'),
+             _('Warn About _Unfillable Squares'),
              None,
              _('Warn about squares made unfillable by a move'),
              self.impossible_implication_cb),
-            ('Tracker', 'tracks', _('_Track additions'),
+            ('Tracker', 'tracks', _('_Track Additions'),
              '<Control>T',
              _('Mark new additions in a separate color so you can keep track of them.'),
              self.tracker_toggle_cb, False),
             ('ToggleToolbar', None, _('Show _Toolbar'), None, None, self.toggle_toolbar_cb, True),
             ('ToggleHighlight', gtk.STOCK_SELECT_COLOR, _('_Highlighter'),
-             None, _('Highlight the current row, column and box'), self.toggle_highlight_cb, False),
-            ('BackgroundGenerator', None, _('Generate new puzzles _while you play'),
-             None,
-             _('Generate new puzzles in the background while you play. This will automatically pause when the game goes into the background.'),
-             self.toggle_generator_cb, True),
+             None, _('Highlight the current row, column and box'), self.toggle_highlight_cb, False)
             ])
 
-        self.edit_actions = gtk.ActionGroup('EditActions')
-        self.edit_actions.add_actions(
-            [('Edit', None, _('_Edit')),
-             ('Undo', gtk.STOCK_UNDO, _('_Undo'), '<Control>z', _('Undo last action'), self.stop_dancer),
-             ('Redo', gtk.STOCK_REDO, _('_Redo'), '<Shift><Control>z', _('Redo last action')),
-             ('Clear', gtk.STOCK_CLEAR, _('_Clear'), '<Control>b', _("Clear entries you've filled in"), self.clear_cb),
-             ('ClearNotes', None, _('Clear _Notes'), None, _("Clear notes and hints"), self.clear_notes_cb),
-             ])
+        self.main_actions.get_action('Undo').set_is_important(True)
+        self.main_actions.get_action('Redo').set_is_important(True)
+        self.main_actions.get_action('ShowPossible').set_is_important(True)
+        self.main_actions.get_action('Tracker').set_is_important(True)
+
         self.uimanager.insert_action_group(self.main_actions, 0)
-        self.uimanager.insert_action_group(self.edit_actions, 0)
         self.uimanager.add_ui_from_string(self.ui)
 
     def setup_undo (self):
         self.cleared = [] # used for Undo memory
         self.cleared_notes = [] # used for Undo memory
         # Set up our UNDO stuff
-        undo_widg = self.edit_actions.get_action('Undo')
-        redo_widg = self.edit_actions.get_action('Redo')
+        undo_widg = self.main_actions.get_action('Undo')
+        redo_widg = self.main_actions.get_action('Redo')
         self.history = Undo.UndoHistoryList(undo_widg, redo_widg)
         for entry in self.gsd.__entries__.values():
             Undo.UndoableGenericWidget(entry, self.history,
-                                       set_method = 'set_value_from_undo',
+                                       set_method = 'set_value_for_undo',
+                                       get_method = 'get_value_for_undo',
                                        pre_change_signal = 'value-about-to-change'
                                        )
             Undo.UndoableGenericWidget(entry, self.history,
-                                       set_method = 'set_notes',
-                                       get_method = 'get_note_text',
+                                       set_method = 'set_notes_for_undo',
+                                       get_method = 'get_notes_for_undo',
                                        signal = 'notes-changed',
-                                       pre_change_signal = 'value-about-to-change',
+                                       pre_change_signal = 'notes-about-to-change',
                                        )
 
     def setup_color (self):
@@ -359,9 +334,6 @@ class UI (gconf_wrapper.GConfWrapper):
         self.vb.show()
         self.game_box.show()
         self.main_area.pack_start(self.game_box, False, padding = 12)
-        self.statusbar = gtk.Statusbar()
-        self.statusbar.show()
-        self.vb.pack_end(self.statusbar, fill = False, expand = False)
         self.w.add(self.vb)
 
     def setup_toggles (self):
@@ -371,8 +343,6 @@ class UI (gconf_wrapper.GConfWrapper):
               self.main_actions.get_action('AlwaysShowPossible')),
              ('show_impossible_implications',
               self.main_actions.get_action('ShowImpossibleImplications')),
-             ('generate_puzzles_in_background',
-              self.main_actions.get_action('BackgroundGenerator')),
              ('show_toolbar',
               self.main_actions.get_action('ToggleToolbar')),
              ('highlight',
@@ -408,6 +378,11 @@ class UI (gconf_wrapper.GConfWrapper):
             self.dancer.stop_dancing()
             delattr(self, 'dancer')
 
+    def start_dancer (self):
+        import dancer
+        self.dancer = dancer.GridDancer(self.gsd)
+        self.dancer.start_dancing()
+
     @simple_debug
     def you_win_callback (self, grid):
         if hasattr(self, 'dancer'):
@@ -417,24 +392,21 @@ class UI (gconf_wrapper.GConfWrapper):
         self.gconf['difficulty'] = self.gconf['difficulty'] + 0.1
         self.timer.finish_timing()
         self.sudoku_tracker.finish_game(self)
-        sublabel = _("You completed the puzzle in %(totalTime)s (%(activeTime)s active)") % {'totalTime': self.timer.total_time_string(),
-        'activeTime': self.timer.active_time_string()
-                }
+        if self.timer.active_time != self.timer.total_time:
+            sublabel = _("You completed the puzzle in %(totalTime)s (%(activeTime)s active).") % {'totalTime': self.timer.total_time_string(),
+            'activeTime': self.timer.active_time_string()
+                    }
+        else:
+            sublabel = _("You completed the puzzle in %(totalTime)s.") % {'totalTime': self.timer.total_time_string()}
         sublabel += "\n"
-        sublabel += ngettext("You got %(n)s hint", "You got %(n)s hints", self.gsd.hints) % {'n':self.gsd.hints}
+        sublabel += ngettext("You got %(n)s hint.", "You got %(n)s hints.", self.gsd.hints) % {'n':self.gsd.hints}
         sublabel += "\n"
         if self.gsd.impossible_hints:
             sublabel += ngettext("You had %(n)s impossibility pointed out.",
                                  "You had %(n)s impossibilities pointed out.",
                                  self.gsd.impossible_hints) % {'n':self.gsd.impossible_hints}
             sublabel += "\n"
-        if self.gsd.auto_fills:
-            sublabel += ngettext("You used the auto-fill %(n)s time",
-                                 "You used the auto-fill %(n)s times",
-                                 self.gsd.auto_fills) % {'n':self.gsd.auto_fills}
-        import dancer
-        self.dancer = dancer.GridDancer(self.gsd)
-        self.dancer.start_dancing()
+        self.start_dancer()
         dialog_extras.show_message(_("You win!"), label = _("You win!"),
                                    sublabel = sublabel
                                    )
@@ -485,6 +457,7 @@ class UI (gconf_wrapper.GConfWrapper):
         self.tracker_ui.reset()
         self.history.clear()
         self.won = False
+        self.old_tracker_view = None
 
     @simple_debug
     def resize_cb (self, widget, event):
@@ -535,36 +508,82 @@ class UI (gconf_wrapper.GConfWrapper):
             self.is_fullscreen = True
 
     @simple_debug
-    def clear_cb (self, *args):
+    def game_reset_cb (self, *args):
         clearer = Undo.UndoableObject(
-            self.do_clear, #action
-            self.undo_clear, #inverse
+            self.do_game_reset, #action
+            self.undo_game_reset, #inverse
             self.history #history
             )
         clearer.perform()
 
-    # add a check to stop the dancer if she is dancing
-    def do_clear (self, *args):
+    def do_game_reset (self, *args):
+        self.gsd.cover_track()
+        self.cleared.append(self.tinfo.save())
         self.cleared.append(self.gsd.reset_grid())
+        self.cleared_notes.append((tracker_info.NO_TRACKER, self.gsd.clear_notes('All')))
+        self.tinfo.reset()
         self.stop_dancer()
 
-    # add a check for finish in the undo to clear
-    def undo_clear (self, *args):
+    def undo_game_reset (self, *args):
+        self.tracker_ui.select_tracker(tracker_info.NO_TRACKER)
         for entry in self.cleared.pop():
             self.gsd.add_value(*entry)
-        if self.gsd.grid.check_for_completeness():
-            self.gsd.emit('puzzle-finished')
+        self.tinfo.load(self.cleared.pop())
+        self.tracker_ui.select_tracker(self.tinfo.current_tracker)
+        self.gsd.show_track()
+        self.undo_clear_notes()
 
-    def clear_notes_cb (self, *args):
+    def clear_top_notes_cb (self, *args):
         clearer = Undo.UndoableObject(
-            lambda *args: self.cleared_notes.append(self.gsd.clear_notes()), #action
-            # clear_notes returns a list of tuples indicating the cleared notes...
-            # (x,y,(top,bottom)) -- this is what we need for undoing
-            lambda *args: [self.gsd.__entries__[t[0], t[1]].set_notes(t[2]) for t in self.cleared_notes.pop()], #inverse
+            lambda *args: self.do_clear_notes('Top'), #action
+            self.undo_clear_notes, #inverse
             self.history
             )
         clearer.perform()
 
+    def clear_bottom_notes_cb (self, *args):
+        clearer = Undo.UndoableObject(
+            lambda *args: self.do_clear_notes('Bottom'), #action
+            self.undo_clear_notes, #inverse
+            self.history
+            )
+        clearer.perform()
+
+    def do_clear_notes(self, side):
+        ''' Clear top, bottom, or all notes - in undoable fashion
+
+        The side argument is used to specify which notes are to be cleared.
+        'Top' - Clear only the top notes
+        'Bottom' - Clear only the bottom notes
+        '''
+        self.cleared_notes.append((self.tinfo.current_tracker, self.gsd.clear_notes(side)))
+        # Turn off auto-hint if the player clears the bottom notes
+        if side == 'Bottom' and self.gconf['always_show_hints']:
+            always_show_hint_wdgt = self.main_actions.get_action('AlwaysShowPossible')
+            always_show_hint_wdgt.activate()
+        # Update the hints...in case we're redoing a clear of them
+        if self.gconf['always_show_hints']:
+            self.gsd.update_all_hints()
+
+    def undo_clear_notes(self):
+        ''' Undo previously cleared notes
+
+        Clearing notes fills the cleared_notes list of notes that were cleared.
+        '''
+        cleared_tracker, cleared_notes = self.cleared_notes.pop()
+        # Change the tracker selection if it was tracking during the clear
+        if cleared_tracker != tracker_info.NO_TRACKER:
+            self.tracker_ui.select_tracker(cleared_tracker)
+        self.gsd.apply_notelist(cleared_notes)
+        # Update the hints...in case we're undoing over top of them
+        if self.gconf['always_show_hints']:
+            self.gsd.update_all_hints()
+        # Redraw the notes
+        self.gsd.update_all_notes()
+        # Make sure we're still dancing if we undo after win
+        if self.gsd.grid.check_for_completeness():
+            self.start_dancer()
+
     @simple_debug
     def show_hint_cb (self, *args):
         self.gsd.show_hint()
@@ -576,55 +595,35 @@ class UI (gconf_wrapper.GConfWrapper):
             self.gsd.update_all_hints()
         else:
             self.gsd.always_show_hints = False
-            self.gsd.clear_hints()
+            self.gsd.clear_notes('AutoHint')
 
     @simple_debug
     def impossible_implication_cb (self, action):
         if action.get_active():
-            self.gsd.show_impossible_implications = True
+            self.gsd.display_impossible_implications()
         else:
-            self.gsd.show_impossible_implications = False
-
-    @simple_debug
-    def auto_fill_cb (self, *args):
-        if not hasattr(self, 'autofilled'):
-            self.autofilled = []
-        if not hasattr(self, 'autofiller'):
-            self.autofiller = Undo.UndoableObject(
-                self.do_auto_fill,
-                self.undo_auto_fill,
-                self.history
-                )
-        self.autofiller.perform()
-
-    def do_auto_fill (self, *args):
-        self.autofilled.append(self.gsd.auto_fill())
-        if self.gconf['always_show_hints']:
-            self.gsd.update_all_hints()
-
-    def undo_auto_fill (self, *args):
-        for entry in self.autofilled.pop():
-            self.gsd.remove(entry[0], entry[1], do_removal = True)
-        if self.gconf['always_show_hints']:
-            self.gsd.update_all_hints()
-
-    @simple_debug
-    def auto_fill_current_square_cb (self, *args):
-        self.gsd.auto_fill_current_entry()
+            self.gsd.hide_impossible_implications()
 
     @simple_debug
     def setup_tracker_interface (self):
-        self.trackers = {}
         self.tracker_ui = TrackerBox(self)
         self.tracker_ui.show_all()
         self.tracker_ui.hide()
+        self.tinfo = tracker_info.TrackerInfo()
+        self.old_tracker_view = None
         self.game_box.add(self.tracker_ui)
 
     @simple_debug
     def tracker_toggle_cb (self, widg):
         if widg.get_active():
+            if self.old_tracker_view:
+                self.tinfo.set_tracker_view(self.old_tracker_view)
+                self.tracker_ui.select_tracker(self.tinfo.current_tracker)
+                self.gsd.show_track()
             self.tracker_ui.show_all()
         else:
+            self.old_tracker_view = self.tinfo.get_tracker_view()
+            self.tracker_ui.hide_tracker_cb(None)
             self.tracker_ui.hide()
 
     @simple_debug
@@ -634,27 +633,6 @@ class UI (gconf_wrapper.GConfWrapper):
         else:
             self.tb.hide()
 
-    def set_statusbar_value (self, status):
-        if not hasattr(self, 'sbid'):
-            self.sbid = self.statusbar.get_context_id('game_info')
-        self.statusbar.pop(self.sbid)
-        self.statusbar.push(self.sbid, status)
-
-
-    def update_statusbar (self, *args):
-        if not self.gsd.grid:
-            self.set_statusbar_value(" ")
-            return True
-
-        puzzle = self.gsd.grid.virgin.to_string()
-        puzzle_diff = self.sudoku_maker.get_difficulty(puzzle)
-
-        tot_string = _("Playing %(difficulty)s puzzle.") % {'difficulty':puzzle_diff.value_string()}
-        tot_string += " " + "(%1.2f)" % puzzle_diff.value
-
-        self.set_statusbar_value(tot_string)
-        return True
-
     def toggle_highlight_cb (self, widg):
         if widg.get_active():
             self.gsd.toggle_highlight(True)
@@ -688,13 +666,6 @@ class UI (gconf_wrapper.GConfWrapper):
                                    label = _("Puzzle Statistics"),
                                    sublabel = information)
 
-    @simple_debug
-    def toggle_generator_cb (self, toggle):
-        if toggle.get_active():
-            self.start_worker_thread()
-        else:
-            self.stop_worker_thread()
-
     @simple_debug
     def autosave (self):
         # this is called on a regular loop and will autosave if we
@@ -738,10 +709,6 @@ class UI (gconf_wrapper.GConfWrapper):
         gp = printing.GamePrinter(self.sudoku_maker, self.gconf)
         gp.run_dialog()
 
-    @simple_debug
-    def generate_puzzle_gui (self, *args):
-        sudoku_generator_gui.GameGenerator(self, self.gconf)
-
 class TrackerBox (gtk.VBox):
 
     @simple_debug
@@ -752,6 +719,8 @@ class TrackerBox (gtk.VBox):
         self.builder.set_translation_domain(DOMAIN)
         self.builder.add_from_file(os.path.join(UI_DIR, 'tracker.ui'))
         self.main_ui = main_ui
+        self.tinfo = tracker_info.TrackerInfo()
+        self.tinfo.ui = self
         self.vb = self.builder.get_object('vbox1')
         self.vb.unparent()
         self.pack_start(self.vb, expand = True, fill = True)
@@ -765,53 +734,79 @@ class TrackerBox (gtk.VBox):
         for tree in self.tracker_model:
             if tree[0] > -1:
                 self.tracker_model.remove(tree.iter)
+        self.tinfo.reset()
+        self.tracker_actions.set_sensitive(False)
 
     @simple_debug
     def setup_tree (self):
-        self.tracker_tree = self.builder.get_object('treeview1')
+        self.tracker_tree = self.builder.get_object('TrackerTreeView')
         self.tracker_model = gtk.ListStore(int, gtk.gdk.Pixbuf, str)
+        self.tracker_model.set_sort_column_id(0, gtk.SORT_ASCENDING)
         self.tracker_tree.set_model(self.tracker_model)
         col1 = gtk.TreeViewColumn("", gtk.CellRendererPixbuf(), pixbuf = 1)
-        col2 = gtk.TreeViewColumn("", gtk.CellRendererText(), text = 2)
+        rend = gtk.CellRendererText()
+        col2 = gtk.TreeViewColumn("", rend, text = 2)
+        col2.set_cell_data_func(rend, self.draw_tracker_name)
         self.tracker_tree.append_column(col2)
         self.tracker_tree.append_column(col1)
         # Our initial row...
-        self.tracker_model.append([-1, None, _('No Tracker')])
+        pixbuf = self.pixbuf_transform_color(
+            STOCK_PIXBUFS['tracks'],
+            (0, 0, 0)
+            )
+        self.tracker_model.append([-1, pixbuf, _('Untracked')])
         self.tracker_tree.get_selection().connect('changed', self.selection_changed_cb)
 
     @simple_debug
     def setup_actions (self):
         self.tracker_actions = gtk.ActionGroup('tracker_actions')
         self.tracker_actions.add_actions(
-            [('Clear',
+            [('Remove',
               gtk.STOCK_CLEAR,
-              _('_Clear Tracker'),
-              None, _('Clear all moves tracked by selected tracker.'),
-              self.clear_cb
+              _('_Remove'),
+              None, _('Delete selected tracker.'),
+              self.remove_tracker_cb
+              ),
+             ('Hide',
+              gtk.STOCK_CLEAR,
+              _('H_ide'),
+              None, _('Hide current tracker entries.'),
+              self.hide_tracker_cb
+              ),
+             ('Apply',
+              gtk.STOCK_APPLY,
+              _('A_pply'),
+              None, _('Apply all tracked values and remove the tracker.'),
+              self.apply_tracker_cb
               ),
-             ('Keep', None,
-              _('_Clear Others'),
-              None,
-              _('Clear all moves not tracked by selected tracker.'),
-              self.keep_cb),
              ]
             )
-        for action, widget_name in [('Clear', 'ClearTrackerButton'),
-                                   ('Keep', 'KeepTrackerButton'),
-                                   ]:
-            a = self.tracker_actions.get_action(action)
-            a.connect_proxy(self.builder.get_object(widget_name))
+        a = self.tracker_actions.get_action('Remove')
+        a.connect_proxy(self.builder.get_object('RemoveTrackerButton'))
+        a = self.tracker_actions.get_action('Hide')
+        a.connect_proxy(self.builder.get_object('HideTrackerButton'))
+        a = self.tracker_actions.get_action('Apply')
+        a.connect_proxy(self.builder.get_object('ApplyTrackerButton'))
         self.builder.get_object('AddTrackerButton').connect('clicked',
                                                           self.add_tracker)
         # Default to insensitive (they only become sensitive once a tracker is added)
         self.tracker_actions.set_sensitive(False)
 
+    def draw_tracker_name(self, column, cell, model, iter):
+        if model.get_value(iter, 0) == self.tinfo.showing_tracker:
+            cell.set_property('weight', pango.WEIGHT_BOLD)
+        else:
+            cell.set_property('weight', pango.WEIGHT_NORMAL)
+
     @simple_debug
-    def add_tracker (self, *args):
-        tracker_id = self.main_ui.gsd.create_tracker()
+    def add_tracker (self, *args, **keys):
+        if keys and keys.has_key('tracker_id'):
+            tracker_id = self.tinfo.create_tracker(keys['tracker_id'])
+        else:
+            tracker_id = self.tinfo.create_tracker()
         pixbuf = self.pixbuf_transform_color(
             STOCK_PIXBUFS['tracks'],
-            self.main_ui.gsd.get_tracker_color(tracker_id),
+            self.tinfo.get_color(tracker_id)
             )
         # select our new tracker
         self.tracker_tree.get_selection().select_iter(
@@ -820,6 +815,7 @@ class TrackerBox (gtk.VBox):
                                   _("Tracker %s") % (tracker_id + 1)]
                                   )
             )
+        self.tinfo.set_tracker(tracker_id)
 
     @simple_debug
     def pixbuf_transform_color (self, pixbuf, color):
@@ -837,10 +833,28 @@ class TrackerBox (gtk.VBox):
                                             pixbuf.get_width(), pixbuf.get_height(), pixbuf.get_rowstride())
 
     @simple_debug
-    def select_tracker (self, tracker_id):
+    def find_tracker (self, tracker_id):
         for row in self.tracker_model:
             if row[0] == tracker_id:
-                self.tracker_tree.get_selection().select_iter(row.iter)
+                return row
+        return None
+
+    @simple_debug
+    def select_tracker (self, tracker_id):
+        track_row = self.find_tracker(tracker_id)
+        if track_row:
+            self.tracker_tree.get_selection().select_iter(track_row.iter)
+            self.tinfo.set_tracker(tracker_id)
+
+    def redraw_row(self, tracker_id):
+        track_row = self.find_tracker(tracker_id)
+        if track_row:
+            self.tracker_model.row_changed(self.tracker_model.get_path(track_row.iter), track_row.iter)
+
+    def set_tracker_action_sense(self, enabled):
+        self.tracker_actions.set_sensitive(True)
+        for action in self.tracker_actions.list_actions():
+            action.set_sensitive(self.tinfo.showing_tracker != tracker_info.NO_TRACKER)
 
     @simple_debug
     def selection_changed_cb (self, selection):
@@ -848,47 +862,130 @@ class TrackerBox (gtk.VBox):
         if itr:
             selected_tracker_id = mod.get_value(itr, 0)
         else:
-            selected_tracker_id = -1
-        # This should be cheap since we don't expect many trackers...
-        # We cycle through each row and toggle it off if it's not
-        # selected; on if it is selected
-        for row in self.tracker_model:
-            tid = row[0]
-            if tid != -1: # -1 == no tracker
-                self.main_ui.gsd.toggle_tracker(tid, tid == selected_tracker_id)
-        self.tracker_actions.set_sensitive(selected_tracker_id != -1)
+            selected_tracker_id = tracker_info.NO_TRACKER
+        if selected_tracker_id != tracker_info.NO_TRACKER:
+            self.main_ui.gsd.cover_track()
+        # Remove the underline on the showing_tracker
+        self.redraw_row(self.tinfo.showing_tracker)
+        self.tinfo.set_tracker(selected_tracker_id)
+        self.set_tracker_action_sense(self.tinfo.showing_tracker != tracker_info.NO_TRACKER)
+        # Show the tracker
+        if selected_tracker_id != tracker_info.NO_TRACKER:
+            self.main_ui.gsd.show_track()
+        self.main_ui.gsd.update_all_notes()
+        if self.main_ui.gconf['always_show_hints']:
+            self.main_ui.gsd.update_all_hints()
 
     @simple_debug
-    def clear_cb (self, action):
+    def remove_tracker_cb (self, action):
         mod, itr = self.tracker_tree.get_selection().get_selected()
         # This should only be called if there is an itr, but we'll
         # double-check just in case.
         if itr:
-            selected_tracker_id = mod.get_value(itr, 0)
-            self.tracker_delete_tracks(selected_tracker_id)
+            clearer = Undo.UndoableObject(
+                self.do_delete_tracker,
+                self.undo_delete_tracker,
+                self.main_ui.history
+                )
+            clearer.perform()
 
     @simple_debug
-    def keep_cb (self, action):
-        mod, itr = self.tracker_tree.get_selection().get_selected()
-        selected_tracker_id = mod.get_value(itr, 0)
-        self.tracker_keep_tracks(selected_tracker_id)
+    def hide_tracker_cb (self, action):
+        hiding_tracker = self.tinfo.showing_tracker
+        self.select_tracker(tracker_info.NO_TRACKER)
+        self.main_ui.gsd.cover_track(True)
+        self.main_ui.gsd.update_all_notes()
+        self.set_tracker_action_sense(False)
+        self.redraw_row(hiding_tracker)
+        self.redraw_row(tracker_info.NO_TRACKER)
 
     @simple_debug
-    def tracker_delete_tracks (self, tracker_id):
-        clearer = Undo.UndoableObject(
-            lambda *args: self.main_ui.cleared.append(self.main_ui.gsd.delete_by_tracker(tracker_id)),
-            lambda *args: [self.main_ui.gsd.add_value(*entry) for entry in self.main_ui.cleared.pop()],
-            self.main_ui.history)
-        clearer.perform()
+    def apply_tracker_cb (self, action):
+        '''Apply Tracker button action
+        '''
+        # Shouldn't be here if no tracker is showing
+        if self.tinfo.showing_tracker == tracker_info.NO_TRACKER:
+            return
+        # Apply the tracker in undo-able fashion
+        applyer = Undo.UndoableObject(
+            self.do_apply_tracker,
+            self.undo_apply_tracker,
+            self.main_ui.history
+            )
+        applyer.perform()
 
-    @simple_debug
-    def tracker_keep_tracks (self, tracker_id):
-        clearer = Undo.UndoableObject(
-            lambda *args: self.main_ui.cleared.append(self.main_ui.gsd.delete_except_for_tracker(tracker_id)),
-            lambda *args: [self.main_ui.gsd.add_value(*entry) for entry in self.main_ui.cleared.pop()],
-            self.main_ui.history)
-        clearer.perform()
+    def do_apply_tracker(self):
+        '''Apply the showing tracker to untracked
 
+        All of the values and notes will be transferred to untracked and
+        the tracker is deleted.
+        '''
+        track_row = self.find_tracker(self.tinfo.showing_tracker)
+        if not track_row:
+            return
+        # Delete the tracker
+        cleared_values, cleared_notes = self.do_delete_tracker(True)
+        # Apply the values
+        for x, y, val, tid in cleared_values:
+            self.main_ui.gsd.set_value(x, y, val)
+        # Then apply the notes
+        self.main_ui.gsd.apply_notelist(cleared_notes, True)
+        # Store the undo counts
+        self.main_ui.cleared.append(len(cleared_values))
+        self.main_ui.cleared_notes.append(len(cleared_notes))
+
+    def undo_apply_tracker(self):
+        '''Undo a previous tracker apply
+
+        The number of cleared values and notes are stored during the apply.
+        The undo is called for each of them, then the tracker delete is
+        undone.
+        '''
+        # Undo all of the applied values and notes
+        value_count = self.main_ui.cleared.pop()
+        note_count = self.main_ui.cleared_notes.pop()
+        count = 0
+        while count < (value_count + note_count):
+            self.main_ui.history.undo()
+            count += 1
+        # Undo the tracker delete
+        self.undo_delete_tracker()
+
+    def do_delete_tracker(self, for_apply = False):
+        '''Delete the current tracker
+        '''
+        track_row = self.find_tracker(self.tinfo.showing_tracker)
+        if not track_row:
+            return
+        ui_row = [track_row[0], track_row[1], track_row[2]]
+        # For the values, store it like (tracker_id, list_of_cleared_values)
+        cleared_values = self.main_ui.gsd.delete_by_tracker()
+        self.main_ui.cleared.append((self.tinfo.showing_tracker, ui_row, cleared_values))
+        # The notes already have tracker info in them, so just store the list
+        cleared_notes = self.main_ui.gsd.clear_notes(tracker = self.tinfo.showing_tracker)
+        self.main_ui.cleared_notes.append(cleared_notes)
+        # Delete it from tracker_info
+        self.hide_tracker_cb(None)
+        self.tracker_model.remove(track_row.iter)
+        self.tinfo.delete_tracker(ui_row[0])
+        # Return all of the data for "Apply Tracker" button
+        if for_apply:
+            return (cleared_values, cleared_notes)
+
+    def undo_delete_tracker(self):
+        '''Undo a tracker delete
+        '''
+        # Values are stored like (tracker_id, list_of_cleared_values)
+        tracker_id, ui_row, cleared_values = self.main_ui.cleared.pop()
+        # Recreate it in tracker_info
+        self.tinfo.create_tracker(tracker_id)
+        # Add it to the tree
+        self.tracker_tree.get_selection().select_iter(self.tracker_model.append(ui_row))
+        # Add all the values
+        for value in cleared_values:
+            self.main_ui.gsd.add_value(*value)
+        # The notes already have tracker info in them, so just store the list
+        self.main_ui.gsd.apply_notelist(self.main_ui.cleared_notes.pop())
 
 def start_game ():
     if options.debug:
